--[[
------------------------
--PubSub: implementation for the subscribe/publish pattern.
-------------------------
]]--

--- a PubSub module responsible for subscription and publishing of events
-- @module PubSub
function PubSub()
  local self = {}
  local events = {}

--- private function for debugging purposes.
-- @function printObjectsListeningToEvent
-- @param event
  local function printObjectsListeningToEvent(event)
    --collectgarbage()
    for k,v in pairs(events[event]) do
      print(k,v)
    end
  end

--- private function - gets the object index in the events table.
-- @function printObjectsListeningToEvent
-- @param obj
-- @param event
-- @return the index if the object exists in the events table, -1 otherwise.
  local function getObjectIndex(obj,event)
    if events == nil or events[event] == nil then return end
    for k,v in pairs(events[event]) do
      if v == obj then return k end
    end
    return -1
  end

  --- private function - unsubscribes an object from a specific event
  -- @function unsubscribeSpecificEvent
  -- @param obj
  -- @param event
  -- @return true if the unsubscription succeeded, false otherwise
  local function unsubscribeSpecificEvent(obj,event)
    local index = getObjectIndex(obj, event)
      if index ~= -1 then
        print("Object "..tostring(obj).." has stopped listening to event "..event)
        table.remove(events[event], index)
        return true
      else
        print("Object "..tostring(obj).." was not listening to event "..event)
        return false
      end
  end
  --- subscribe to event
  -- @function subscribeEvent
  -- @param obj
  -- @param event
  -- @return true if the unsubscription succeeded, false otherwise
  function self.subscribeEvent(obj, event)
    --priority is for now inactive
    local params = validation.new{ obj = obj, event = event}
    params.obj:validate(validation.type("table"))
    params.event:validate(validation.string.len{ min = 2 })

    if not params.valid then
      print("Event listener not added. error:", params.fields, params.errors)
      return false
    end

     --collectgarbage()

    local flag = false
    if events[event] == nil then
      events[event] = {}
      setmetatable(events[event], { __mode = 'v' })
    end

    --look for the object
    for k,v in pairs(events[event]) do
      if v == obj then
        print("Object "..tostring(obj).." is already listening to the event "..event)
        return true
      end
    end

    if obj["on"..event] == nil then
      print("Object "..tostring(obj).." does not have a method named on"..event)
      return false
    end

    table.insert(events[event], obj)
     print("Object "..tostring(obj).." is now listening to the event "..event)
     return true
  end

--setPriority method not relevant.
--[[
  function self.setPriority(obj, event, priority)
    local params = validation.new{ obj =obj, event = event, priority = priority }
    params.obj:validate(validation.type("table"))
    params.event:validate(validation.string.len{ min = 2 })
    params.priority:validate(validation.number.min(0))

    if not params.valid then print("Priority not set. error:", params.fields, params.errors) return false end
    if events == nil or events[event] == nil then print("There are no objects listening to event "..event) return false end

    local index = getObjectIndex(obj, event)
    if index == -1 then print("Priority not set. Object "..obj.." is not listening to event "..event) return false
  else
      events[event][index] = nil
      if priority ~= nil then
        table.insert(events[event], priority, obj)
      else
       table.insert(events[event], obj)
      end
      return true
    end
  end
  --]]
  --- publish event with parameters
  -- @function publishEvent
  -- @param event
  -- @param ... list of parameters to pass to the onEvent method to be triggered
  function self.publishEvent(event, ...)
    local numOfFunctionsTriggered = 0
    local params = validation.new{event = event}
    params.event:validate(validation.string.len{ min = 2 })
    if not params.valid then print("Could not dispatch event. error:", params.fields, params.errors) return false end
    --collectgarbage()
    if events == nil or events[event] == nil then print("There are no objects listening to event "..event) return false end

    --remove this
    for k,v in pairs(events[event]) do
      print(k,v)
      if v["on"..event] ~= nil then
        v["on"..event](...)
        numOfFunctionsTriggered = numOfFunctionsTriggered +1
      end
    end
    if numOfFunctionsTriggered > 0 then
      return true
    else print("There are no objects listening to event "..event)
      --release event
      events[event]=nil
      return false
    end
  end

  --- unsubscribe event
  -- @function unsubscribeEvent
  -- @param obj
  -- @param event optional. If not used, will unsubscribe the objects from all the events it is subscribed to.
  -- @return true if the unsubscription succeeded, false otherwise
  function self.unsubscribeEvent(obj, event)
    local params = validation.new{ obj =obj, event = event, priority = priority }
    local success = false
    params.obj:validate(validation.typeOrNil("table"))
     params.event:validate(validation.string.len{ min = 2 })

    if not params.valid then print("Listener not removed. Error:", params.fields, params.errors) return false end
    if events == nil or events[event] == nil then print("There are no objects listening to event "..event) return false end

    success = unsubscribeSpecificEvent(obj, event)
    if success== true and events[event] ~= nil and #events[event]== 0 then events[event] = nil end
    return success
  end

  function self.unsubscribeFromAllEvents(obj)
    local success = false
    for _,currentEvent in pairs(events) do
        local result = unsubscribeSpecificEvent(obj, event)
        if result == true then success = true end
    end
    return success
  end


  return self
end



--self.unsubscribeFromAllEvents(obj)

--Initiate (singleton)

PubSub = PubSub()
print("initiating singleton pub sub"..tostring(PubSub))

--return PubSub


